/* hello.c
 *
 * Testing routines.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"

static char usage[]  =
"Usage: hello <mode>\n"
"Modes:\n"
"   g1    Generative training 1 step, B8\n"
"   t12   Generative training, T12\n"
"   b8    Generative training, B8\n"
"   vet   Generative training, SEEKAVET\n"
"   fb    Forward-Backward algorithm using probabilities and scales\n"
"   diag  Writes DIAG hmm file, for hmmemit test\n"
"   adfh  Compare forward-backward with matlab implementation\n"
"   wri   Write sequence\n"
"   ass   Assert-Forward-Backward\n"
"   ass1  Assert-Forward-Backward with default HMM\n"
"   post  Posterior decoding using the ADFH motif";

static void TestGen1();
static void TestGenT12();
static void TestGenB8();
static void TestGenVet();
static void TestMleForBack();
static void TestWriteHMM();
static void TestADFH();
static void TestPosteriorADFH();
static void TestWriteSeq();
static void TestAssertFB();
static void TestAssertFB1();

int main(int argc, char **argv)
{
  time_t       starttime;
  int          minutes;

  if (argc < 2) {printf("%s\n", usage); exit(0);}

  starttime = time((time_t *) NULL);

  if      (strcmp(argv[1], "g1" ) == 0) TestGen1();
  else if (strcmp(argv[1], "t12") == 0) TestGenT12();
  else if (strcmp(argv[1], "b8" ) == 0) TestGenB8();
  else if (strcmp(argv[1], "vet") == 0) TestGenVet();
  else if (strcmp(argv[1], "fb" ) == 0) TestMleForBack();
  else if (strcmp(argv[1], "diag")== 0) TestWriteHMM();
  else if (strcmp(argv[1], "adfh")== 0) TestADFH();
  else if (strcmp(argv[1], "wri") == 0) TestWriteSeq();
  else if (strcmp(argv[1], "ass") == 0) TestAssertFB();
  else if (strcmp(argv[1], "ass1")== 0) TestAssertFB1();
  else if (strcmp(argv[1], "post")== 0) TestPosteriorADFH();
  else {printf("%s\n", usage); exit(0);}

  minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
  printf("Elasped time is %d:%d\n", minutes / 60, minutes % 60);
  return 0;
}

static void TestGen1()
{
  char *seqfile         = "CuTest/yeast_seq_B.fa";
  char *hmmfile         = "CuTest/trn_B8.hmm";
  int   setlength       = 8;
  int   maxiter         = 10;
  int   verbose         = 1;

  //RunGenTrain
  int nseqs = 0;
  SQINFO *sqinfo;
  SetAlphabet(hmmAMINO);
  unsigned char **dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo);

  //GenTrain
  float randomseq[MAXABET];     /* null sequence model emission         */
  float p1;                     /* null sequence model p1 transition    */
  struct p7prior_s *prior = P7DefaultPrior();
  struct plan7_s *hmm = AllocPlan7(setlength);
  ZeroPlan7(hmm);
  //P7DefaultNullModel(randomseq, &p1);
  P7UniformNullModel(hmm); //Plan7SetNullModel(hmm, randomseq, p1);
  P7PriorifyHMM(hmm, prior);
  Plan7GlobalConfig(hmm);

  //TrainOptimize
  int idx;
  float *wt = (float *) MallocOrDie(nseqs * sizeof(float));
  for (idx = 0; idx < nseqs; idx++) wt[idx] = 1.0;
  struct plan7_s *newhmm;
  P7Logoddsify(hmm, 0);
  PrintHMM(hmm);
  if ((newhmm = AllocPlan7(hmm->M)) == NULL) /* counts-based HMM */
    Die("Failed to allocate for a new HMM\n");
  ZeroPlan7(newhmm);
  P7UniformNullModel(newhmm); //Plan7SetNullModel(newhmm, hmm->null, hmm->p1);
  float score = TrainCountSuffStat(nseqs, dsqs, sqinfo, wt, hmm, newhmm);
  P7PriorifyHMM(newhmm, prior);
  Plan7GlobalConfig(newhmm);
  FreePlan7(hmm);
  hmm = newhmm;

  P7Logoddsify(hmm, 0);
  PrintHMM(hmm);
  if ((newhmm = AllocPlan7(hmm->M)) == NULL) /* counts-based HMM */
    Die("Failed to allocate for a new HMM\n");
  ZeroPlan7(newhmm);
  P7UniformNullModel(newhmm); //Plan7SetNullModel(newhmm, hmm->null, hmm->p1);
  score = TrainCountSuffStat(nseqs, dsqs, sqinfo, wt, hmm, newhmm);
  P7PriorifyHMM(newhmm, prior);
  Plan7GlobalConfig(newhmm);
  FreePlan7(hmm);
  hmm = newhmm;

  P7Logoddsify(hmm, 0);
  PrintHMM(hmm);

  P7FreePrior(prior);
  FreePlan7(hmm);
  free(wt);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(dsqs); free(sqinfo);
}

static void TestGenT12()
{
  RunGenTrain("CuTest/yeast_seq_T.fa", "CuTest/trn_T12.hmm", 12, 20, 10, 20., 530., 1);
}

static void TestGenB8()
{
  RunGenTrain("CuTest/yeast_seq_B.fa", "CuTest/trn_B8.hmm", 8, 20, 10, 20., 530., 1);
}

static void TestGenVet()
{
  RunGenTrain("CuTest/seekavet.fa", "CuTest/seekavet.hmm", 8, 30, 10, 20., 530., 1);
}

#define PHMMLEN 8

static char motif[] = "seekavet";

static void TestWriteHMM()
{
  int mat[PHMMLEN][20] = {
    { 5, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 5, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 0, 0, 5, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 0, 0, 0, 0, 5, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 0, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 0, 0, 0 }
  };

  SetAlphabet(hmmAMINO);
  float randomseq[MAXABET];     /* null sequence model emission         */
  float p1;                     /* null sequence model p1 transition    */
  struct p7prior_s *prior = P7DefaultPrior();
  struct plan7_s *hmm;
  if ((hmm = AllocPlan7(PHMMLEN)) == NULL) /* counts-based HMM */
    Die("Failed to allocate for a new HMM\n");
  ZeroPlan7(hmm);
  P7DefaultNullModel(randomseq, &p1);
  Plan7SetNullModel(hmm, randomseq, p1);

  unsigned char *dsq = DigitizeSequence(motif, strlen(motif));
  int i, x;
  for (i = 1; i <= PHMMLEN; i++)
    for (x = 0; x < 20; x++) hmm->mat[i][x] += (float) mat[i-1][x];
  /* insert emission kept zero and will become the prior emission */

  P7PriorifyHMM(hmm, prior);
  P7RetractWing(hmm);
  Plan7GlobalConfig(hmm);
  PrintHMM(hmm);

  /* save as text file (I don't know why binary file cannot be read by HMMER) */
  FILE *hmmfp;
  if ((hmmfp = fopen("diag.hmm", "w")) == NULL)
    Die("Failed to open diag.hmm for writing hmm result\n");
  WriteAscHMM(hmmfp, hmm);
  fclose(hmmfp);

  P7FreePrior(prior);
  FreePlan7(hmm);
}

static void AssignEmission(struct plan7_s *hmm)
{
  unsigned char *dsq = DigitizeSequence(motif, strlen(motif));
  int i;
  for (i = 1; i <= strlen(motif); i++) hmm->mat[i][dsq[i]] += 10;
}

static void TestMleForBack()
{
  char *seqfile = "CuTest/seekavet.fa";
  int setlength = 8;
  struct mledpmat_s *fwd, *bck;

  int nseqs = 0, idx;
  SQINFO *sqinfo = NULL;
  char **rseq;
  unsigned char **dsqs = NULL;
  float sclprob;

  /* load sequence */
  SetAlphabet(hmmAMINO);
  dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo);

  int i, k, x;
  float randomseq[MAXABET];     /* null sequence model emission         */
  float p1;                     /* null sequence model p1 transition    */
  struct p7prior_s *prior = P7SimplePrior(); /* P7DefaultPrior(); */
  struct plan7_s *hmm = AllocPlan7(setlength);
  ZeroPlan7(hmm);
  P7DefaultNullModel(randomseq, &p1);
  Plan7SetNullModel(hmm, randomseq, p1);
  //AssignEmission(hmm);
  P7PriorifyHMM(hmm, prior);
  P7RetractWing(hmm);
  Plan7GlobalConfig(hmm);
  PrintHMM(hmm);
  sclprob = MleForward(dsqs[0], sqinfo[0].len, hmm, &fwd);
  printf("Forward scaled prob = %e\n", sclprob);
  PrintMleDpMat(fwd, sqinfo[0].len, hmm->M);
  float prob_b = MleBackward(dsqs[0], sqinfo[0].len, hmm, fwd->scl, &bck);
  printf("Backward scaled prob = %e\n", prob_b);
  PrintMleDpMat(bck, sqinfo[0].len, hmm->M);

  printf("Posterior:\n");
  for (i = 0; i <= sqinfo[0].len; i++)
    {
      float tot = fwd->xmx[i][XMN] * bck->xmx[i][XMN]
                + fwd->xmx[i][XMC] * bck->xmx[i][XMC];
      for (k = 1; k <= hmm->M; k++)
        tot += fwd->mmx[i][k] * bck->mmx[i][k] + fwd->imx[i][k] * bck->imx[i][k];
      printf("N %f, C %f\n", fwd->xmx[i][XMN] * bck->xmx[i][XMN] / tot,
                             fwd->xmx[i][XMC] * bck->xmx[i][XMC] / tot);
    }

  k = 2;
  struct plan7_s *newhmm;       /* re-estimated hmm */
  if ((newhmm = AllocPlan7(hmm->M)) == NULL) /* counts-based HMM */
    Die("Failed to allocate for a new HMM\n");
  ZeroPlan7(newhmm);
  Plan7SetNullModel(newhmm, hmm->null, hmm->p1);
  MleForbackCount(hmm, dsqs[0], sqinfo[0].len, sclprob, 1., fwd, bck, newhmm);
  printf("M1->M2 %f, M1->I1 %f, M1->D2 %f\n", newhmm->t[k][TMM], newhmm->t[k][TMI], newhmm->t[k][TMD]);
  /* Seemingly correct answer: M1->M2 1.054208, M1->I1 0.033127, M1->D2 0.021463 */
  FreePlan7(newhmm);

  P7FreePrior(prior);
  FreePlan7(hmm);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(dsqs); free(sqinfo);
}

static void PrintDpAs1Mat(struct mledpmat_s *fwd, int L, int M)
{
  int k, i;
  printf("\nN  "); for (i=0; i<=L; i++) printf("%.4f ", fwd->xmx[i][XMN]);
  printf("\nC  "); for (i=0; i<=L; i++) printf("%.4f ", fwd->xmx[i][XMC]);
  for (k = 1; k <= M; k++) {
    printf("\nM%d ", k); for (i=0; i<=L; i++) printf("%.4f ", fwd->mmx[i][k]);
  }
  for (k = 1; k < M; k++) {
    printf("\nI%d ", k); for (i=0; i<=L; i++) printf("%.4f ", fwd->imx[i][k]);
  }
  printf("\nB  "); for (i=0; i<=L; i++) printf("%.4f ", fwd->xmx[i][XMB]);
  printf("\nE  "); for (i=0; i<=L; i++) printf("%.4f ", fwd->xmx[i][XME]);
  for (k = 1; k < M; k++) {
    printf("\nD%d ", k); for (i=0; i<=L; i++) printf("%.4f ", fwd->dmx[i][k]);
  }
}

static char ADFH_FASTA[] = "rstADFHvwy";

static void TestADFH()
{
  int L = strlen(ADFH_FASTA);
  int setlength = 4;
  struct mledpmat_s *fwd, *bck;
  float sclprob;

  SetAlphabet(hmmAMINO);
  unsigned char *dsq = DigitizeSequence(ADFH_FASTA, L);

  struct p7prior_s *prior = P7SimplePrior(); /* P7DefaultPrior(); */
  struct plan7_s *hmm = AllocPlan7(setlength);
  ZeroPlan7(hmm);
  Plan7SetNullModel(hmm, prior->i[0], 350. / 351.);
  P7PriorifyHMM(hmm, prior);
  P7RetractWing(hmm);
  Plan7GlobalConfig(hmm);
  PrintHMM(hmm);
  sclprob = MleForward(dsq, L, hmm, &fwd);
  float prob_b = MleBackward(dsq, L, hmm, fwd->scl, &bck);

  int i;
  printf("\nSC ");
  for (i = 0; i <= L; i++) printf("%.4f ", fwd->scl[i]);
  printf("\nFS %e", sclprob);
  PrintDpAs1Mat(fwd, L, hmm->M);
  printf("\nBS %e", prob_b);
  PrintDpAs1Mat(bck, L, hmm->M);
  printf("\n");

  P7FreePrior(prior);
  FreePlan7(hmm);
  free(dsq);
}

static void TestPosteriorADFH()
{
  int L, i, x;
  int setlength = 4;
  float *begin;
  float *end;

  L = strlen(ADFH_FASTA);
  SetAlphabet(hmmAMINO);
  unsigned char *dsq = DigitizeSequence(ADFH_FASTA, L);

  struct p7prior_s *prior = P7SimplePrior(); /* P7DefaultPrior(); */
  struct plan7_s *hmm = AllocPlan7(setlength);
  ZeroPlan7(hmm);
  Plan7SetNullModel(hmm, prior->i[0], 350. / 351.);

  for (i = 1; i <= setlength; i++) {
    hmm->mat[i][2*i-2] += 5; hmm->mat[i][2*i-1] += 5;
  }

  P7PriorifyHMM(hmm, prior);
  P7RetractWing(hmm);
  Plan7GlobalConfig(hmm);
  PrintHMM(hmm);

  begin = (float *) MallocOrDie(sizeof(float) * (1+L));
  end   = (float *) MallocOrDie(sizeof(float) * (1+L));
  PosteriorBE(dsq, L, hmm, begin, end);
  for (i = 0; i <= L; i++) printf("%.0f ", floor(10. * begin[i]));
  printf("\n");
  for (i = 0; i <= L; i++) printf("%.0f ", floor(10. * end[i]));
  printf("\n");
  free(begin); free(end);

  P7FreePrior(prior);
  FreePlan7(hmm);
  free(dsq);
}

static void TestWriteSeq()
{
  int nseqs = 0, idx;
  SQINFO *sqinfo = NULL;
  unsigned char **dsqs = LoadDsqs("CuTest/yeast_seq_O.fa", &nseqs, &sqinfo);
  FILE *seqfp = fopen("hello.fa", "w");

  printf("Sequence length: %d\n", sqinfo[1].len);
  char *seq = DedigitizeSequence(dsqs[1], sqinfo[1].len);
  WriteSeq(seqfp, SQFILE_FASTA, seq, sqinfo+1);
  fclose(seqfp);

  free(seq);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(sqinfo); free(dsqs);
}

static void TestFB(struct plan7_s *hmm, unsigned char **dsq, SQINFO *sqinfo)
{
  int idx = 0, i = 1;
  float sclprob = 0., loglik = 0.;
  struct mledpmat_s *fwd; //dpmatrix_s *fwd;
  struct mledpmat_s *bck; //dpmatrix_s *bck;
  struct plan7_s *newhmm;

  // Prepare newhmm
  if ((newhmm = AllocPlan7(hmm->M)) == NULL) /* counts-based HMM */
    Die("Failed to allocate for a new HMM\n");
  ZeroPlan7(newhmm);
  Plan7SetNullModel(newhmm, hmm->null, hmm->p1);

  // Run forward-backward
  sclprob = MleForward(dsq[idx], sqinfo[idx].len, hmm, &fwd);
  MleBackward(dsq[idx], sqinfo[idx].len, hmm, fwd->scl, &bck);
  loglik += MleFwdScore(fwd, sqinfo[idx].len, hmm->M, sclprob);

  MleForbackCount(hmm, dsq[idx], sqinfo[idx].len, sclprob, 1.0,
                  fwd, bck, newhmm);

  printf("dsq[%d]: t10=%.2f, sclprob=%f\nscl: ",idx,newhmm->t[1][0],sclprob);
  for (i = 1; i < sqinfo[idx].len; i++) printf("%.2f ",fwd->scl[i]);
  printf("\nFWD:\n"); PrintMleDpMat(fwd, sqinfo[idx].len, hmm->M);
  printf("BCK:\n"); PrintMleDpMat(bck, sqinfo[idx].len, hmm->M);

  // Free
  FreePlan7(newhmm);
  FreeMleDpMat(fwd); FreeMleDpMat(bck);
}

static void TestAssertFB()
{
  int nseqs = 0, idx = 0;
  SQINFO *sqinfo = NULL;
  unsigned char **dsq = LoadDsqs("assert.fa", &nseqs, &sqinfo);
  struct plan7_s *hmm = LoadHMM("assert.hmm");

  PrintHMM(hmm);
  TestFB(hmm, dsq, sqinfo);

  FreePlan7(hmm);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsq[idx], &(sqinfo[idx]));
  free(sqinfo); free(dsq);
}

static void TestAssertFB1()
{
  int nseqs = 0, idx = 0;
  SQINFO *sqinfo = NULL;
  unsigned char **dsq = LoadDsqs("assert.fa", &nseqs, &sqinfo);

  //Random / default HMM
  struct plan7_s *hmm = NULL;
  struct p7prior_s *prior = P7DefaultPrior();
  float randomseq[MAXABET];     /* null sequence model emission         */
  float p1;                     /* null sequence model p1 transition    */
  P7DefaultNullModel(randomseq, &p1);
  hmm = P7Init(4, prior, randomseq, p1, 0, NULL, NULL);

  TestFB(hmm, dsq, sqinfo);

  FreePlan7(hmm);
  P7FreePrior(prior);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsq[idx], &(sqinfo[idx]));
  free(sqinfo); free(dsq);
}

static void TestAssertFB2()
{
  int nseqs = 0, idx = 0;
  SQINFO *sqinfo = NULL;
  unsigned char **dsq = LoadDsqs("assert2.fa", &nseqs, &sqinfo);

  //Random / default HMM
  struct plan7_s *hmm = NULL;
  struct p7prior_s *prior = P7DefaultPrior();
  float randomseq[MAXABET];     /* null sequence model emission         */
  float p1;                     /* null sequence model p1 transition    */
  P7DefaultNullModel(randomseq, &p1);
  hmm = P7Init(4, prior, randomseq, p1, 0, NULL, NULL);

  TestFB(hmm, dsq, sqinfo);

  FreePlan7(hmm);
  P7FreePrior(prior);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsq[idx], &(sqinfo[idx]));
  free(sqinfo); free(dsq);
}
