/* subcv.c
 *
 * Cross validation.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"

static char usage[]  =
"Usage: subcv [-options] <seqfile> <label1> <label2>...\n"
"Options:\n"
"   -l <length>       : profile HMM (match-states) length (default 10)\n"
"   -r <learn-rate>   : learning rate (default 0.01)\n"
"   -i <iterations>   : maximum iterations (default 20)\n"
"   -I <iterations>   : maximum iterations, discriminative (default 40)\n"
"   -n <fold>         : n-fold cross validation (default 4)\n"
"   -p <permutations> : number of repeats with random permutation (default 1)\n"
"   -b <begin>        : number of random beginnings, the first being average (default 1)\n"
"   -m <max-insert>   : maximum insertions (default 20.0)\n"
"   -a <avg-length>   : average length for background (default 530.0)\n"
"   -s <seed>         : random seed (default by random)\n"
"   -o <outfile>      : output result file (default ../result/A_BC_D17.result)\n"
"   -v                : verbose (default OFF)\n";

static struct opt_s OPTIONS[] = {
  { "-l", TRUE, sqdARG_INT    },
  { "-r", TRUE, sqdARG_FLOAT  },
  { "-i", TRUE, sqdARG_INT    },
  { "-I", TRUE, sqdARG_INT    },
  { "-n", TRUE, sqdARG_INT    },
  { "-p", TRUE, sqdARG_INT    },
  { "-b", TRUE, sqdARG_INT    },
  { "-m", TRUE, sqdARG_FLOAT  },
  { "-a", TRUE, sqdARG_FLOAT  },
  { "-s", TRUE, sqdARG_INT    },
  { "-v", TRUE, sqdARG_NONE   },
  { "-o", TRUE, sqdARG_STRING },
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

const int   DEFAULT_PHMM_LEN    = 10;
const float DEFAULT_LEARN_RATE  = 0.01;
const int   MAXITER             = 20;
const int   MAXITER1            = 40;
const int   DEFAULT_NFOLD       = 4;
const int   DEFAULT_NREPEAT     = 1;
const int   DEFAULT_NINIT       = 1;
const float DEFAULT_INSLIMIT    = 20.0;
const float DEFAULT_AVGLEN      = 530.0;

static char *ResultFilename(int ncate, char **labels, int setlength);

int
main(int argc, char **argv)
{
  int   setlength       = DEFAULT_PHMM_LEN;
  float learnrt         = DEFAULT_LEARN_RATE;
  int   maxiter         = MAXITER;
  int   maxiter1        = MAXITER1;
  int   nfold           = DEFAULT_NFOLD;
  int   nrepeat         = DEFAULT_NREPEAT;
  int   ninit           = DEFAULT_NINIT;
  float inslimit        = DEFAULT_INSLIMIT;
  float avglen          = DEFAULT_AVGLEN;
  int   seed            = (int) time((time_t *) NULL); /* "random" seed */
  char *resultfile      = NULL;
  int   verbose         = 0;
  char *optname;                /* name of option found by Getopt()         */
  char *optarg;                 /* argument found by Getopt()               */
  int   optind;                 /* index in argv[]                          */

  char        *seqfile;
  int          ncate;
  char       **labels;
  time_t       starttime;
  int          minutes;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
    if      (strcmp(optname, "-l") == 0) setlength     = atoi(optarg);
    else if (strcmp(optname, "-r") == 0) learnrt       = atof(optarg);
    else if (strcmp(optname, "-i") == 0) maxiter       = atoi(optarg);
    else if (strcmp(optname, "-I") == 0) maxiter1      = atoi(optarg);
    else if (strcmp(optname, "-n") == 0) nfold         = atoi(optarg);
    else if (strcmp(optname, "-p") == 0) nrepeat       = atoi(optarg);
    else if (strcmp(optname, "-b") == 0) ninit         = atoi(optarg);
    else if (strcmp(optname, "-m") == 0) inslimit      = atof(optarg);
    else if (strcmp(optname, "-a") == 0) avglen        = atof(optarg);
    else if (strcmp(optname, "-s") == 0) seed          = atoi(optarg);
    else if (strcmp(optname, "-o") == 0) resultfile    = optarg;
    else if (strcmp(optname, "-v") == 0) verbose       = 1;
    else if (strcmp(optname, "-h") == 0)
      {
        printf("subcv: cross validation\n\n%s\n", usage);
        exit(0);
      }
  }
  if (argc - optind < 3)
    Die("Incorrect number of arguments.\n%s\n", usage);

  seqfile = argv[optind++];
  ncate = argc - optind;
  labels = argv + optind;
  if (NULL == resultfile) resultfile = ResultFilename(ncate,labels,setlength);
  starttime = time((time_t *) NULL);

  /* Main routine begins */
  int *catenseq = NULL;
  SQINFO **catesqinfo = NULL;
  unsigned char ***catedsqs = LoadLabeledDsqs(seqfile, ncate, labels, &catenseq,
                                              &catesqinfo);
  RunCrossValidate(ncate, catenseq, catedsqs, catesqinfo, nfold, nrepeat,
                   setlength, learnrt, ninit, maxiter, maxiter1, inslimit,
                   avglen, seed, verbose, "", resultfile);//FIXME
  FreeCateDsqs(ncate, catenseq, catedsqs, catesqinfo);
  /* Main routine ends */

  minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
  printf("%d-fold CV %d repeats finished in %d:%d results written to %s\n",
         nfold, nrepeat, minutes / 60, minutes % 60, resultfile);
  return 0;
}

#define RESULT_FILE_BUF_LEN 80

static char result_file_buffer[RESULT_FILE_BUF_LEN] = "../result/";

/* Return result filename by adding motif length and extension ".result"
 * Note that this implementation is buffer-based,
 * so caller need not free returned name.
 */
static char *
ResultFilename(int ncate, char **labels, int setlength)
{
  int k;
  /*strcpy(result_file_buffer, "../data/");*/
  /*result_file_buffer[0] = 0;*/
  strcat(result_file_buffer, labels[0]);
  for (k = 1; k < ncate; k++)
    {
      strcat(result_file_buffer, "_");
      strcat(result_file_buffer, labels[k]);
    }
  sprintf(result_file_buffer + strlen(result_file_buffer), "%02d", setlength);
  strcat(result_file_buffer, ".result");
  return result_file_buffer;
}
