/* subfeat.c
 *
 * Generate feature file from profile HMMs.
 * Features can be either the log likelihood ratio of HMMs vs background model,
 * or the log likelihood of background and HMMs.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"

static char usage[]  =
"Usage: subfeat [-options] <seqfile> <featfile> <hmmprefix1> <hmmprefix2>...\n"
"Options:\n"
"   -n <num-motif> : number of motifs, using random masking (default 10)\n"
"   --LLR          : output LLR (default)\n"
"   --loglik       : output background and HMM log likelihood\n"
"   -v <level>     : verbose level (default 0, least information)\n";

static struct opt_s OPTIONS[] = {
  { "-n",       TRUE, sqdARG_INT    },
  { "--LLR",    TRUE, sqdARG_NONE   },
  { "--loglik", FALSE,sqdARG_NONE   },
  { "-v",       TRUE, sqdARG_INT    }
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

const int   DEFAULT_NMOTIF      = 10;

static char **HmmFilenames(int nhmmprefix, char **hmmprefix, int nmotif, int *ret_nhmm);
static void RunGenFeatures(char *seqfile, int nhmm, char **hmmfiles, char *featfile, int outformat, int verbose);

int
main(int argc, char **argv)
{
  char        *optname;                         /* name of option from Getopt() */
  char        *optarg;                          /* argument from Getopt()       */
  int          optind;                          /* index in argv[]              */
  int          verbose          = 0;
  int          outformat        = 0;            /* 0:LLR, 1:loglike */
  int          nmotif           = DEFAULT_NMOTIF;
  int          nhmmprefix       = 0;
  char        *seqfile          = NULL;
  char        *featfile         = NULL;
  char       **hmmprefix        = NULL;
  char       **hmmfiles         = NULL;
  int          nhmm             = 0;
  int          ihmm             = 0;
  time_t       starttime        = 0;
  int          minutes          = 0;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
    if      (strcmp(optname, "-v") == 0) verbose = atoi(optarg);
    else if (strcmp(optname, "-n") == 0) nmotif = atoi(optarg);
    else if (strcmp(optname, "--loglik") == 0) outformat = 1;
    /*else if (strcmp(optname, "-s") == 0) seed = atoi(optarg);*/
    else if (strcmp(optname, "-h") == 0)
      {
        printf("subfeat: generate feature file from profile HMMs\n\n%s\n", usage);
        exit(0);
      }
  }
  if (argc - optind < 3)
    Die("Incorrect number of arguments.\n%s\n", usage);

  seqfile = argv[optind++];
  featfile = argv[optind++];
  hmmprefix = argv + optind;
  nhmmprefix = argc - optind;
  hmmfiles = HmmFilenames(nhmmprefix, hmmprefix, nmotif, &nhmm);

  starttime = time((time_t *) NULL);
  RunGenFeatures(seqfile, nhmm, hmmfiles, featfile, outformat, verbose);
  minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
  printf("Finished in %d:%d\n", minutes / 60, minutes % 60);

  for (ihmm = 0; ihmm < nhmm; ihmm++) free(hmmfiles[ihmm]);
  free(hmmfiles);
  return 0;
}

/***********************************************
 * Filenames
 ***********************************************/

static char **
HmmFilenames(int nhmmprefix, char **hmmprefix, int nmotif, int *ret_nhmm)
{
  int nhmm = nhmmprefix * nmotif;
  char **hmmfiles = (char **) MallocOrDie(sizeof(char *) * nhmm);
  int ihmm = 0, i, j;
  for (i = 0; i < nhmmprefix; i++)
    for (j = 0; j < nmotif; j++)
      {
        hmmfiles[ihmm] = (char *) MallocOrDie(strlen(hmmprefix[i]) + 7);
        sprintf(hmmfiles[ihmm], "%s%02d.hmm", hmmprefix[i], j);
        ihmm++;
      }
  assert(ihmm == nhmm);
  *ret_nhmm = nhmm;
  return hmmfiles;
}

/***********************************************
 * Likelihood ratio
 ***********************************************/

/* Calculate the likelihood of the null model.
 * Should be in a shared module if used by other programs.
 */
static float
CalNullModelScore(float null[MAXABET], float p1, unsigned char *dsq, int L)
{
  int i;
  float score = L * sreLOG2(p1) + sreLOG2(1.0 - p1);
  for (i = 1; i <= L; i++) score += sreLOG2(null[dsq[i]]);
  return score;
}

static double **
GenFeatures(int nseqs, unsigned char **dsqs, SQINFO *sqinfo, int nhmm,
            char **hmmfiles, int outformat, int verbose)
{
  int idx, ihmm;
  double **feats = (double **) MallocOrDie(nseqs * sizeof(double *));
  for (idx = 0; idx < nseqs; idx++)
    {
      if (outformat==1)
        feats[idx] = (double *) MallocOrDie(sizeof(double) * (nhmm+1));
      else
        feats[idx] = (double *) MallocOrDie(sizeof(double) * nhmm);
    }
  for (ihmm = 0; ihmm < nhmm; ihmm++)
    {
      if (verbose) printf("Processing HMM file %s...\n", hmmfiles[ihmm]);
      struct plan7_s *hmm = LoadHMM(hmmfiles[ihmm]);
      if (verbose > 1)
        {
          int x;
          PrintHMM(hmm);
          printf("p1=%.3f, null= ", hmm->p1);
          for (x = 0; x < Alphabet_size; x++) printf("%f ", hmm->null[x]);
          printf("\n");
        }
      for (idx = 0; idx < nseqs; idx++)
        {
          struct mledpmat_s *fwd;
          float scl_score = MleForward(dsqs[idx], sqinfo[idx].len, hmm, &fwd);
          double loglik = MleFwdScore(fwd, sqinfo[idx].len, hmm->M, scl_score);
          double loglik0 = CalNullModelScore(hmm->null, hmm->p1, dsqs[idx],
                                            sqinfo[idx].len);
          if (verbose > 1) printf("%f - %f, ", loglik, loglik0);
          if (outformat == 1) /* write log likelihood */
            {
              feats[idx][ihmm+1] = loglik;
              if (ihmm == 0) { feats[idx][0] = loglik0; }
              else if (fabs(loglik0-feats[idx][0])>1e-5) printf("bg:%.3f ",loglik0);
            }
          else { feats[idx][ihmm] = loglik - loglik0; } /* write LLR */
          FreeMleDpMat(fwd);
        }
      if (verbose > 1) printf("\n");
      FreePlan7(hmm);
    }
  return feats;
}

static void
RunGenFeatures(char *seqfile, int nhmm, char **hmmfiles, char *featfile,
               int outformat, int verbose)
{
  int             idx;
  int             nseqs         = 0;
  unsigned char **dsqs          = NULL;
  SQINFO         *sqinfo        = NULL;
  double        **feats         = NULL;

  SetAlphabet(hmmAMINO);
  if (verbose) printf("Loading sequence %s...\n", seqfile);
  dsqs = LoadDsqs(seqfile, &nseqs, &sqinfo); /* load sequence and digitize */
  feats = GenFeatures(nseqs, dsqs, sqinfo, nhmm, hmmfiles, outformat, verbose);
  if (verbose) printf("Saving features %s...\n", featfile);
  SaveFeatFile(nseqs, (outformat==1) ? nhmm+1 : nhmm, feats, featfile);
  for (idx = 0; idx < nseqs; idx++) free(feats[idx]);
  free(feats);
  for (idx = 0; idx < nseqs; idx++) FreeDsqinfo(dsqs[idx], &(sqinfo[idx]));
  free(dsqs); free(sqinfo);
}
