/* sublearn.c
 *
 * Learn a Plan 7 Profile HMM from given sequence file by
 * generative training (Baum-Welch) or discriminative (Extended Baum-Welch)
 * training algorithm. Optimal path found by Viterbi algorithm is displayed
 * to help examine learned profile.
 */

#include "config.h"             /* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "structs.h"            /* data structures, macros, #define's   */
#include "funcs.h"              /* function declarations                */
#include "globals.h"            /* alphabet global variables            */
#include "squid.h"              /* general sequence analysis library    */
#include "subloc.h"

static char usage[]  =
"Usage: sublearn [-options] <seqfile>...\n"
"Options:\n"
"   -l <length>       : profile HMM (match-states) length (default 10)\n"
"   -i <iterations>   : maximum iterations (default 20)\n"
"   -b <begin>        : number of random beginnings, the first being average (default 1)\n"
"   -m <max-insert>   : maximum insertions (default 20.0)\n"
"   -a <avg-length>   : average length for background (default 530.0)\n"
"   -s <seed>         : random seed (default by random)\n"
"   -o <outfile>      : output HMM file OUTFILE.hmm and OUTFILE.hmm.txt\n"
"   -v <level>        : verbose level (default 0, least information)\n"; /*
"   -r <learn-rate>   : learning rate (default 0.01)\n"
"   -I <iterations>   : maximum iterations, discriminative (default 40)\n" */

static struct opt_s OPTIONS[] = {
  { "-l", TRUE, sqdARG_INT    },
  { "-i", TRUE, sqdARG_INT    },
  { "-b", TRUE, sqdARG_INT    },
  { "-m", TRUE, sqdARG_FLOAT  },
  { "-a", TRUE, sqdARG_FLOAT  },
  { "-s", TRUE, sqdARG_INT    },
  { "-v", TRUE, sqdARG_INT    },
  { "-o", TRUE, sqdARG_STRING }/*,
  { "-r", TRUE, sqdARG_FLOAT  },
  { "-I", TRUE, sqdARG_INT    },*/
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

const int   DEFAULT_PHMM_LEN    = 10;
const float DEFAULT_LEARN_RATE  = 0.01;
const int   MAXITER             = 20;
const int   MAXITER1            = 40;
const int   DEFAULT_NINIT       = 1;
const float DEFAULT_INSLIMIT    = 20.0;
const float DEFAULT_AVGLEN      = 530.0;

int
main(int argc, char **argv)
{
  int   setlength       = DEFAULT_PHMM_LEN;
  float learnrt         = DEFAULT_LEARN_RATE;
  int   maxiter         = MAXITER;
  int   maxiter1        = MAXITER1;
  int   ninit           = DEFAULT_NINIT;
  float inslimit        = DEFAULT_INSLIMIT;
  float avglen          = DEFAULT_AVGLEN;
  int   seed            = (int) time((time_t *) NULL); /* "random" seed */
  char *outfile         = "sublearnout";
  int   verbose         = 0;
  char *optname;                /* name of option found by Getopt()         */
  char *optarg;                 /* argument found by Getopt()               */
  int   optind;                 /* index in argv[]                          */

  char        *seqfile;
  int          ncate;
  char       **labels;
  time_t       starttime;
  int          minutes;

  while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
    if      (strcmp(optname, "-l") == 0) setlength     = atoi(optarg);
    else if (strcmp(optname, "-i") == 0) maxiter       = atoi(optarg);
    /*else if (strcmp(optname, "-I") == 0) maxiter1      = atoi(optarg);
    else if (strcmp(optname, "-r") == 0) learnrt       = atof(optarg);*/
    else if (strcmp(optname, "-b") == 0) ninit         = atoi(optarg);
    else if (strcmp(optname, "-m") == 0) inslimit      = atof(optarg);
    else if (strcmp(optname, "-a") == 0) avglen        = atof(optarg);
    else if (strcmp(optname, "-s") == 0) seed          = atoi(optarg);
    else if (strcmp(optname, "-o") == 0) outfile       = optarg;
    else if (strcmp(optname, "-v") == 0) verbose       = atoi(optarg);
    else if (strcmp(optname, "-h") == 0)
      {
        printf("sublearn: learn a HMM\n\n%s\n", usage);
        exit(0);
      }
  }
  if (argc - optind < 1)
    Die("Incorrect number of arguments.\n%s\n", usage);

  seqfile = argv[optind++];
  ncate = argc - optind;
  labels = argv + optind;

  starttime = time((time_t *) NULL);
  RunGenTrain(seqfile, outfile, setlength, maxiter, ninit, inslimit, avglen, verbose);
  //RunMaskGenTrain
  minutes = ((int) difftime(time((time_t *) NULL), starttime) + 30) / 60;
  printf("Finished in %d:%d\n",
         minutes / 60, minutes % 60);
  return 0;
}
